# Created by msinghal at 09/10/23

import os
from pathlib import Path
import pytest
import pickle
from langchain.embeddings.elasticsearch import ElasticsearchEmbeddings
from singulr_client.singulr_tracer import SingulrTracer
from langchain.callbacks.tracers.schemas import Run


# @pytest.fixture
def load_run_examples() -> Run:
    file_path = str(Path(__file__).parents[2] / "examples/run_object_anonymize_chain.pkl")
    with open(file_path, 'r') as f:
        run = pickle.load(f)
    return run


def test_singulr_trace_generation() -> None:
    pass



# Created by msinghal at 12/09/23
import logging
import logging.config
import colorlog

# Default log path and log format constants
DEFAULT_LOG_PATH = "/tmp"
LOG_FORMAT = '%(asctime)s.%(msecs)03dZ %(levelname)s %(filename)s:%(lineno)d %(message)s'
DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'

# Create a color formatter
COLOR_LOG_FORMAT = colorlog.ColoredFormatter(
    LOG_FORMAT,
    log_colors={
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "bold_red",
    },
    secondary_log_colors={},
    style="%",
)

# Configuration for logging to a file
log_file_config = {
    'version': 1,
    'formatters': {
        'genai_log_formatter': {
            'class': 'logging.Formatter',
            'datefmt': DATE_FORMAT,
            'format': LOG_FORMAT
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'INFO',
        },
        'info_file_handler': {
            'class': 'logging.FileHandler',
            'filename': '',
            'mode': 'a',
            'formatter': 'genai_log_formatter'
        }
    },
    'loggers': {
        'singulr_gen_ai': {
            'handlers': ['info_file_handler']
        }
    },
    'root': {
        'level': 'INFO',
        'handlers': []
    },
}

# Configuration for logging to the console
stream_config = {
    'version': 1,
    'formatters': {
        'genai_log_formatter': {
            'class': 'logging.Formatter',
            'datefmt': DATE_FORMAT,
            'format': LOG_FORMAT
        }
    },
    'handlers': {
        'info_handler': {
            'class': 'logging.StreamHandler',
            'formatter': 'genai_log_formatter'
        }
    },
    'loggers': {
        'app_gen_ai': {
            'handlers': ['info_handler']
        }
    },
    'root': {
        'level': 'INFO',
        'handlers': []
    },
}


# Function to configure the logger
def configure_logger(folder_path: str, is_console_logging: bool, is_debug_enabled: bool) -> None:
    """
    Args:
        folder_path:
        is_console_logging:
        is_debug_enabled:

    Returns:
    """
    if not is_console_logging:
        log_info_file = folder_path + "/latest.log"
        log_file_config['handlers']['info_file_handler']['filename'] = log_info_file
        logging.config.dictConfig(log_file_config)
    else:
        logging.config.dictConfig(stream_config)
    logging.captureWarnings(True)
    if is_debug_enabled:
        logging.root.setLevel(logging.DEBUG)


# Logger for 'singulr-sdk' module
singulr_sdk = logging.getLogger("singulr-sdk")



# Created by msinghal at 12/09/23
import abc
from enum import Enum
from typing import Type, Dict, Any

from typing_extensions import TypedDict
import attr

from pydantic import BaseModel


class AttributeType(str, Enum):
    """
    Enum representing attribute types.
    """
    UNKNOWN = "UNKNOWN"
    PRIMITIVE_STRING = "PRIMITIVE_STRING"
    PRIMITIVE_INT = "PRIMITIVE_INT"
    LLM_INVOCATION = "LLM_INVOCATION"
    PROMPT_TEMPLATE = "PROMPT_TEMPLATE"
    LLM_STATS = "LLM_STATS"

    def __str__(self) -> str:
        return str(self.value)


class Attribute(metaclass=abc.ABCMeta):
    """
    Base class for attribute objects.
    """
    @abc.abstractmethod
    def get_key(self) -> str:
        """
        Get the key associated with the attribute.

        Returns:
            str: The attribute key.
        """
        pass

    @abc.abstractmethod
    def get_type(self) -> AttributeType:
        """
        Get the attribute type.

        Returns:
            AttributeType: The attribute type.
        """
        pass


class BaseAttribute(Attribute):
    """
    Base class for attribute objects with a key and type.
    """
    def __init__(self, key: str, type: AttributeType):
        self.key = key
        self.type = type

    def get_key(self) -> str:
        return self.key

    def get_type(self) -> AttributeType:
        return self.type

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the attribute to a dictionary.

        Returns:
            Dict[str, Any]: The attribute represented as a dictionary.
        """
        return {
            "key": self.key,
            "type": self.type.value,
        }


class PrimitiveAttribute(BaseAttribute, metaclass=abc.ABCMeta):
    """
    Base class for primitive attribute objects.
    """
    def __init__(self, key: str, type: AttributeType):
        super().__init__(key, type)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "key": self.key,
            "type": self.type.value,
        }


class StringAttribute(PrimitiveAttribute):
    """
    Represents a string attribute.

    Attributes:
        key (str): The unique identifier for the attribute.
        value (str): The string value of the attribute.
    """
    def __init__(self, key: str, value: str):
        super().__init__(key, AttributeType.PRIMITIVE_STRING)
        self.value = value

    def to_json(self) -> Dict[str, Any]:
        return {
            "@type": "string",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }

    def to_dict(self) -> Dict[str, Any]:
        return {
            "@type": "string",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }


class IntAttribute(PrimitiveAttribute):
    """
    Represents an integer attribute.

    Attributes:
        key (str): The unique identifier for the attribute.
        value (int): The integer value of the attribute.
    """
    def __init__(self, key: str, value: int):
        super().__init__(key, AttributeType.PRIMITIVE_INT)
        self.value = value

    def to_json(self) -> Dict[str, Any]:
        return {
            "@type": "int",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }

    def to_dict(self) -> Dict[str, Any]:
        return {
            "@type": "int",
            "key": self.key,
            "type": self.type.value,
            "value": self.value
        }


if __name__ == '__main__':
    s = IntAttribute(key="execution_order", value=1)
    debug = True



# Created by msinghal at 11/09/23
from enum import Enum
from typing import List, Dict, Optional, Any

class StatusCode(str, Enum):
    """
    Enumeration for status codes.

    Attributes:
        SUCCESS (str): Represents a successful status.
        ERROR (str): Represents an error status.
    """

    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)


class NetworkInfo:
    """
    Represents network information.

    Args:
        host_name (str, optional): The host name. Default is "fin01.example.com".
        ip_address (str, optional): The IP address. Default is "192.168.2.5".
        subnet (str, optional): The subnet. Default is "192.168.2.0/24".
    """

    def __init__(self, host_name: str = "fin01.example.com", ip_address: str = "192.168.2.5",
                 subnet: str = "192.168.2.0/24"):
        self.host_name = host_name
        self.ip_address = ip_address
        self.subnet = subnet

    def to_dict(self) -> Dict[str, any]:
        """
        Serialize the NetworkInfo object to a dictionary.

        Returns:
            Dict[str, any]: A dictionary representation of the object.
        """
        return {
            "host_name": self.host_name,
            "ip_address": self.ip_address,
            "subnet": self.subnet
        }


class NetworkInterface:
    """
    Represents a network interface.

    Args:
        name (str, optional): The name of the network interface. Default is "en0".
        mac (str, optional): The MAC address. Default is "bc:d0:74:0e:e8:2c".
    """

    def __init__(self, name: str = "en0", mac: str = "bc:d0:74:0e:e8:2c"):
        self.name = name
        self.mac = mac
        self.network = NetworkInfo()

    def to_dict(self) -> Dict[str, any]:
        """
        Serialize the NetworkInterface object to a dictionary.

        Returns:
            Dict[str, any]: A dictionary representation of the object.
        """
        return {
            'name': self.name,
            'mac': self.mac,
            'network': self.network.to_dict() if self.network else None
        }

    def get_name(self) -> str:
        """Get the name of the network interface."""
        return self.name

    def set_name(self, name: str) -> None:
        """Set the name of the network interface."""
        self.name = name

    def get_mac(self) -> str:
        """Get the MAC address of the network interface."""
        return self.mac

    def set_mac(self, mac: str) -> None:
        """Set the MAC address of the network interface."""
        self.mac = mac

    def get_network(self) -> NetworkInfo:
        """Get the network information associated with the network interface."""
        return self.network

    def set_network(self, network: NetworkInfo) -> None:
        """Set the network information associated with the network interface."""
        self.network = network

class MachineInfo:
    """
    Represents information about the machine.

    Args:
        operating_system (str, optional): The operating system. Default is "Linux".
        os_version (str, optional): The OS version. Default is "Ubuntu 20.04".
        total_memory (int, optional): Total memory in megabytes. Default is 8192.
        cpu_info (str, optional): CPU information. Default is "Intel Core i7".
        num_cores (int, optional): Number of CPU cores. Default is 4.
        architecture (str, optional): Machine architecture. Default is "x86_64".
        network_interfaces (List[NetworkInterface], optional): List of network interfaces. Default is None.
    """

    def __init__(self, operating_system: str = "Linux", os_version: str = "Ubuntu 20.04", total_memory: int = 8192,
                 cpu_info: str = "Intel Core i7", num_cores: int = 4, architecture: str = "x86_64",
                 network_interfaces: Optional[List[NetworkInterface]] = None):
        self.operating_system = operating_system
        self.os_version = os_version
        self.total_memory = total_memory
        self.cpu_info = cpu_info
        self.num_cores = num_cores
        self.architecture = architecture
        self.network_interfaces = network_interfaces if network_interfaces is not None else [NetworkInterface()]

    def to_dict(self) -> Dict[str, any]:
        """
        Serialize the MachineInfo object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "operating_system": self.operating_system,
            "os_version": self.os_version,
            "total_memory": self.total_memory,
            "cpu_info": self.cpu_info,
            "num_cores": self.num_cores,
            "architecture": self.architecture,
            "network_interfaces": [network_interface.to_dict() for network_interface in self.network_interfaces]
        }


class EnvironmentContext:
    """
    Represents context information about the environment.

    Args:
        environment_variables (Dict[str, str], optional): Dictionary of environment variables. Default is an empty dictionary.
        command_line_args (List[str], optional): List of command line arguments. Default is an empty list.
        log_files (List[str], optional): List of log files. Default is an empty list.
    """

    def __init__(self, environment_variables: Optional[Dict[str, str]] = None, command_line_args: Optional[List[str]] = None, log_files: Optional[List[str]] = None):
        self.environment_variables = environment_variables if environment_variables is not None else {}
        self.command_line_args = command_line_args if command_line_args is not None else []
        self.log_files = log_files if log_files is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the EnvironmentContext object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "environment_variables": self.environment_variables,
            "command_line_args": self.command_line_args,
            "log_files": self.log_files
        }


class ProcessInfo:
    """
    Represents information about a process.

    Args:
        process_path (str, optional): The path to the process. Default is "python chat_bot_app.py".
        process_id (int, optional): The process ID. Default is 10000.
        status (str, optional): The status of the process. Default is "running".
        listen_ports (List[int], optional): List of listening ports. Default is an empty list.
        environment_context (EnvironmentContext, optional): The environment context associated with the process. Default is None.
    """

    def __init__(self, process_path: str = "python chat_bot_app.py", process_id: int = 10000, status: str = "running", listen_ports: Optional[List[int]] = None,
                 environment_context: Optional[EnvironmentContext] = None):
        self.process_path = process_path
        self.process_id = process_id
        self.status = status
        self.listen_ports = listen_ports if listen_ports is not None else []
        self.environment_context = environment_context if environment_context is not None else EnvironmentContext()

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the ProcessInfo object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "process_path": self.process_path,
            "process_id": self.process_id,
            "status": self.status,
            "listen_ports": self.listen_ports,
            "environment_context": self.environment_context.to_dict()
        }


class RuntimeEnvironment:
    """
    Represents information about the runtime environment.

    Args:
        lang_runtime (str, optional): The language runtime. Default is "python".
        lang_runtime_version (str, optional): The version of the language runtime. Default is "3.11.4".
        entry_point (str, optional): The entry point for the application. Default is "main.py".
        virtual_environment (str, optional): The virtual environment used. Default is "CONDA".
        environment (str, optional): The environment name. Default is "llm-security".
    """

    def __init__(self, lang_runtime: str = "python", lang_runtime_version: str = "3.11.4", entry_point: str = "main.py", virtual_environment: str = "CONDA", environment: str = "llm-security"):
        self.lang_runtime = lang_runtime
        self.lang_runtime_version = lang_runtime_version
        self.entry_point = entry_point
        self.virtual_environment = virtual_environment
        self.environment = environment

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the RuntimeEnvironment object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "lang_runtime": self.lang_runtime,
            "lang_runtime_version": self.lang_runtime_version,
            "entry_point": self.entry_point,
            "virtual_environment": self.virtual_environment,
            "environment": self.environment
        }


class SecurityContext:
    """
    Represents security context information.

    Args:
        user_id (str, optional): The user ID. Default is "root".
        group_id (str, optional): The group ID. Default is "wheel".
    """

    def __init__(self, user_id: str = "root", group_id: str = "wheel"):
        self.user_id = user_id
        self.group_id = group_id

    def to_dict(self) -> Dict[str, str]:
        """
        Serialize the SecurityContext object to a dictionary.

        Returns:
            Dict[str, str]: A dictionary representation of the object.
        """
        return {
            "user_id": self.user_id,
            "group_id": self.group_id
        }


class RequestSecurityContext:
    """
    Represents request security context information.

    Args:
        user_id (str, optional): The user ID. Default is "amit".
        group_id (str, optional): The group ID. Default is "engineering".
    """

    def __init__(self, user_id: str = "amit", group_id: str = "engineering"):
        self.user_id = user_id
        self.group_id = group_id

    def to_dict(self) -> Dict[str, str]:
        """
        Serialize the RequestSecurityContext object to a dictionary.

        Returns:
            Dict[str, str]: A dictionary representation of the object.
        """
        return {
            "user_id": self.user_id,
            "group_id": self.group_id
        }


class FileSystemInfo:
    """
    Represents file system information.

    Args:
        working_directory (str, optional): The working directory. Default is "/apps/app1".
        total_disk_space_mbs (int, optional): Total disk space in megabytes. Default is 102400.
        file_system_paths (List[str], optional): List of file system paths. Default is an empty list.
    """

    def __init__(self, working_directory: str = "/apps/app1", total_disk_space_mbs: int = 102400, file_system_paths: Optional[List[str]] = None):
        self.working_directory = working_directory
        self.total_disk_space_mbs = total_disk_space_mbs
        self.file_system_paths = file_system_paths if file_system_paths is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the FileSystemInfo object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "working_directory": self.working_directory,
            "total_disk_space_mbs": self.total_disk_space_mbs,
            "file_system_paths": self.file_system_paths
        }


class DeploymentInfo:
    """
    Represents deployment information.

    Args:
        name (str, optional): The name of the deployment. Default is "SampleDeployment".
        version (str, optional): The version of the deployment. Default is "1.0.0".
        deployment_environment (str, optional): The deployment environment. Default is "Production".
        metadata (List[Dict[str, str]], optional): List of metadata items. Default is an empty list.
    """

    def __init__(self, name: str = "SampleDeployment", version: str = "1.0.0", deployment_environment: str = "Production", metadata: Optional[List[Dict[str, str]]] = None):
        self.name = name
        self.version = version
        self.deployment_environment = deployment_environment
        self.metadata = metadata if metadata is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the DeploymentInfo object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "name": self.name,
            "version": self.version,
            "deployment_environment": self.deployment_environment,
            "metadata": self.metadata
        }


class FrameworkInfo:
    """
    Represents framework information.

    Args:
        used_frameworks (List[str], optional): List of used frameworks. Default is an empty list.
        dependencies (List[str], optional): List of dependencies. Default is an empty list.
    """

    def __init__(self, used_frameworks: Optional[List[str]] = None, dependencies: Optional[List[str]] = None):
        self.used_frameworks = used_frameworks if used_frameworks is not None else []
        self.dependencies = dependencies if dependencies is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the FrameworkInfo object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "used_frameworks": self.used_frameworks,
            "dependencies": self.dependencies
        }


from typing import List, Dict, Optional, Any


class AppFrameworkMetadata:
    def __init__(self, key: str = "frameworkKey1", value: str = "frameworkValue1", tags: Optional[List[str]] = None):
        """
        Initialize AppFrameworkMetadata.

        Args:
            key (str, optional): The key for the metadata. Default is "frameworkKey1".
            value (str, optional): The value for the metadata. Default is "frameworkValue1".
            tags (List[str], optional): A list of tags for the metadata. Default is an empty list.
        """
        self.key = key
        self.value = value
        self.tags = tags if tags is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the AppFrameworkMetadata object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "key": self.key,
            "value": self.value,
            "tags": self.tags
        }


class ApplicationFramework:
    def __init__(self, app_framework: str = "LANG_CHAIN", version: str = "1.0", app_framework_metadata: Optional[List[AppFrameworkMetadata]] = None):
        """
        Initialize ApplicationFramework.

        Args:
            app_framework (str, optional): The application framework name. Default is "LANG_CHAIN".
            version (str, optional): The version of the application framework. Default is "1.0".
            app_framework_metadata (List[AppFrameworkMetadata], optional): A list of AppFrameworkMetadata objects. Default is an empty list.
        """
        self.app_framework = app_framework
        self.version = version
        self.app_framework_metadata = app_framework_metadata if app_framework_metadata is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the ApplicationFramework object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "app_framework": self.app_framework,
            "version": self.version,
            "app_framework_metadata": [metadata.to_dict() for metadata in self.app_framework_metadata]
        }


class Metadata:
    def __init__(self, key: str = "k1", value: str = "value1", tags: Optional[List[str]] = None):
        """
        Initialize Metadata.

        Args:
            key (str, optional): The key for the metadata. Default is "k1".
            value (str, optional): The value for the metadata. Default is "value1".
            tags (List[str], optional): A list of tags for the metadata. Default is an empty list.
        """
        self.key = key
        self.value = value
        self.tags = tags if tags is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the Metadata object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "key": self.key,
            "value": self.value,
            "tags": self.tags
        }


class ApplicationMetadata:
    def __init__(self, name: str = "app1", version: str = "1.0", description: str = "", owner: str = "madan", metadata: Optional[List[Metadata]] = None):
        """
        Initialize ApplicationMetadata.

        Args:
            name (str, optional): The name of the application. Default is "app1".
            version (str, optional): The version of the application. Default is "1.0".
            description (str, optional): A description of the application. Default is an empty string.
            owner (str, optional): The owner of the application. Default is "madan".
            metadata (List[Metadata], optional): A list of Metadata objects. Default is an empty list.
        """
        self.name = name
        self.version = version
        self.description = description
        self.owner = owner
        self.metadata = metadata if metadata is not None else []

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the ApplicationMetadata object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "owner": self.owner,
            "metadata": [metadata.to_dict() for metadata in self.metadata]
        }




class Environment:
    def __init__(self, machine: Optional[MachineInfo] = None, networks: Optional[List[NetworkInfo]] = None,
                 process: Optional[ProcessInfo] = None, runtime_environment: Optional[RuntimeEnvironment] = None,
                 security_context: Optional[SecurityContext] = None, request_security_context: Optional[RequestSecurityContext] = None,
                 file_system: Optional[FileSystemInfo] = None, deployment: Optional[DeploymentInfo] = None,
                 framework: Optional[FrameworkInfo] = None, application_framework: Optional[ApplicationFramework] = None,
                 application_metadata: Optional[ApplicationMetadata] = None):
        """
        Initialize Environment.

        Args:
            machine (MachineInfo, optional): Information about the machine. Default is None.
            networks (List[NetworkInfo], optional): List of network information. Default is a list containing one NetworkInfo object.
            process (ProcessInfo, optional): Information about the process. Default is None.
            runtime_environment (RuntimeEnvironment, optional): Information about the runtime environment. Default is None.
            security_context (SecurityContext, optional): Security context information. Default is None.
            request_security_context (RequestSecurityContext, optional): Request security context information. Default is None.
            file_system (FileSystemInfo, optional): File system information. Default is None.
            deployment (DeploymentInfo, optional): Deployment information. Default is None.
            framework (FrameworkInfo, optional): Framework information. Default is None.
            application_framework (ApplicationFramework, optional): Application framework information. Default is None.
            application_metadata (ApplicationMetadata, optional): Application metadata. Default is None.
        """
        self.machine = machine if machine else MachineInfo()
        self.networks = networks if networks else [NetworkInfo()]
        self.process = process if process else ProcessInfo()
        self.runtime_environment = runtime_environment if runtime_environment else RuntimeEnvironment()
        self.security_context = security_context if security_context else SecurityContext()
        self.request_security_context = request_security_context if request_security_context else RequestSecurityContext()
        self.file_system = file_system if file_system else FileSystemInfo()
        self.deployment = deployment if deployment else DeploymentInfo()
        self.framework = framework if framework else FrameworkInfo()
        self.application_framework = application_framework if application_framework else ApplicationFramework()
        self.application_metadata = application_metadata if application_metadata else ApplicationMetadata()

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the Environment object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "machine": self.machine.to_dict() if self.machine is not None else None,
            "networks": [network.to_dict() for network in self.networks],
            "process": self.process.to_dict() if self.process is not None else None,
            "runtime_environment": self.runtime_environment.to_dict() if self.runtime_environment is not None else None,
            "security_context": self.security_context.to_dict() if self.security_context is not None else None,
            "request_security_context": self.request_security_context.to_dict() if self.request_security_context else None,
            "file_system": self.file_system.to_dict() if self.file_system is not None else None,
            "deployment": self.deployment.to_dict() if self.deployment is not None else None,
            "framework": self.framework.to_dict() if self.framework is not None else None,
            "application_framework": self.application_framework.to_dict() if self.application_framework is not None else None,
            "application_metadata": self.application_metadata.to_dict() if self.application_metadata is not None else None
        }

    def add_attribute(self, key: str, value: Any) -> None:
        """
        Add an attribute to the environment.

        Args:
            key (str): The key of the attribute.
            value (Any): The value of the attribute.

        Returns:
            None
        """
        if self.attributes is None:
            self.attributes = {}
        self.attributes[key] = value



# Created by msinghal at 12/09/23
from enum import Enum
from typing import Optional, Dict, Union

class ContentType(str, Enum):
    """
    Enum representing content types.
    """
    GENERATION = "GENERATION"
    TEXT_DOCUMENT = "TEXT_DOCUMENT"

class DocumentType(Enum):
    """
    Enum representing document types.
    """
    TEXT = "TEXT"
    ADDRESSABLE_CTX_DOCUMENT = "ADDRESSABLE_CTX_DOCUMENT"
class KeyedContent:
    """
    Represents content with a key, content, and content type.

    Attributes:
        key (str): The unique identifier for the content.
        content (str): The content data.
        content_type (ContentType): The content type.
    """

    def __init__(self, key: str, content: str, content_type: ContentType):
        self.key: str = key
        self.content: str = content
        self.content_type: ContentType = content_type

    def __str__(self) -> str:
        return str(self.value)

    def to_json(self) -> dict:
        return self.value

    def to_dict(self) -> Dict[str, Union[str, ContentType]]:
        """
        Convert the content to a dictionary.

        Returns:
            Dict[str, Union[str, ContentType]]: The content represented as a dictionary.
        """
        return {
            "key": self.key,
            "content_type": self.content_type.value,
            "content": self.content
        }

class GenerationDocument(KeyedContent):
    """
    Represents a generation document with additional generation information.

    Attributes:
        key (str): The unique identifier for the content.
        content (str): The content data.
    """

    def __init__(self, key: str, content: str):
        super().__init__(key, content, ContentType.GENERATION)
        self.generation_info: GenerationInfo = GenerationInfo()

    def to_dict(self) -> Dict[str, Union[str, ContentType]]:
        return {
            "@type": "generation_content",
            "key": self.key,
            "content_type": self.content_type.value,
            "content": self.content
        }

class GenerationInfo:
    """
    Represents information about content generation.

    Attributes:
        finish_reason (Optional[str]): The reason for finishing content generation.
    """

    def __init__(self, finish_reason: Optional[str] = None):
        self.finish_reason: Optional[str] = finish_reason

    def get_finish_reason(self) -> Optional[str]:
        return self.finish_reason

class TextDocument(KeyedContent):
    """
    Represents a text document with optional metadata.

    Attributes:
        key (str): The unique identifier for the content.
        content (str): The content data.
    """

    def __init__(self, key: str, content: str):
        super().__init__(key, content, ContentType.TEXT_DOCUMENT)
        self.document_type: DocumentType = DocumentType.TEXT
        self.metadata: DocumentMetadata = None

    def to_dict(self) -> Dict[str, Union[str, ContentType, DocumentType, Dict[str, Optional[Union[str, int]]]]]:
        return {
            "@type": "text_content",
            "key": self.key,
            "content_type": self.content_type.value,
            "content": self.content,
            "document_type": self.document_type.value,
            "metadata": self.metadata.to_dict() if self.metadata else None
        }

class DocumentMetadata:
    """
    Represents metadata associated with a document.

    Attributes:
        source (Optional[str]): The source of the document.
        pointer (Optional[int]): The pointer associated with the document.
    """

    def __init__(self):
        self.source: Optional[str] = None
        self.pointer: Optional[int] = None

    def get_source(self) -> Optional[str]:
        return self.source

    def get_pointer(self) -> Optional[int]:
        return self.pointer

    def to_dict(self) -> Dict[str, Optional[Union[str, int]]]:
        """
        Convert the metadata to a dictionary.

        Returns:
            Dict[str, Optional[Union[str, int]]]: The metadata represented as a dictionary.
        """
        metadata_dict = {
            "source": self.source,
            "pointer": self.pointer
        }
        # Remove fields with None values
        metadata_dict = {k: v for k, v in metadata_dict.items() if v is not None}
        return metadata_dict





# Created by msinghal at 12/09/23
import platform  # Module for platform information
import os
import socket
import psutil  # Modules for OS, networking, and process information
from functools import lru_cache  # Importing LRU cache decorator
import json
import hashlib
import time
from typing import Any, List, Dict

from langchain.callbacks.tracers.schemas import Run

def _hash_id(s: str) -> str:
    """
    Compute an MD5 hash of a string and return the first 16 characters of the hash.

    Args:
        s (str): The input string to be hashed.

    Returns:
        str: The first 16 characters of the MD5 hash of the input string.
    """
    return hashlib.md5(s.encode("utf-8")).hexdigest()[:16]

def _serialize_io(run_inputs: Dict[str, Any]) -> Dict[str, Any]:
    """
    Serialize input data into a dictionary.

    Args:
        run_inputs (dict): A dictionary containing input data.

    Returns:
        dict: A serialized dictionary containing the input data.
    """
    from google.protobuf.json_format import MessageToJson
    from google.protobuf.message import Message

    serialized_inputs = {}
    for key, value in run_inputs.items():
        if isinstance(value, Message):
            serialized_inputs[key] = MessageToJson(value)
        elif key == "input_documents":
            serialized_inputs.update(
                {f"input_document_{i}": json.dumps(doc) for i, doc in enumerate(value)}
            )
        else:
            serialized_inputs[key] = value
    return serialized_inputs

def _fallback_serialize(obj: Any) -> str:
    """
    Serialize a non-serializable object to a string.

    Args:
        obj (Any): The object to be serialized.

    Returns:
        str: A string representation of the object.
    """
    try:
        return f"<<non-serializable: {type(obj).__qualname__}>>"
    except Exception:
        return "<<non-serializable>>"

def _safe_serialize(obj: Dict[str, Any]) -> str:
    """
    Safely serialize a dictionary to a JSON string.

    Args:
        obj (dict): The dictionary to be serialized.

    Returns:
        str: A JSON string representation of the dictionary.
    """
    try:
        return json.dumps(
            obj,
            skipkeys=True,
            default=_fallback_serialize,
        )
    except Exception:
        return "{}"

class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'to_dict') and callable(obj.to_dict):
            return obj.to_dict()
        elif isinstance(obj, list):
            return [self.default(item) for item in obj]
        return super().default(obj)

def generate_trace_id(run: Run) -> str:
    """
    Generate a trace ID based on the Run object and a timestamp.

    Args:
        run (Run): The Run object to be used for generating the trace ID.

    Returns:
        str: The generated trace ID as a string.
    """
    json_str = _safe_serialize(run)
    timestamp = str(time.time())  # Add a timestamp (current time) to the input string
    input_str = json_str + timestamp
    return _hash_id(input_str)



# Created by msinghal at 04/10/23
from __future__ import annotations
import os
import pandas as pd
import time
from datetime import datetime
from dotenv import load_dotenv, find_dotenv
_ = load_dotenv(find_dotenv())
if os.environ.get('LOG_FOLDER'):
    folder_path = os.environ.get('LOG_FOLDER')
else:
    folder_path = "/tmp"

# set logger
from singulr_client.common import log_config
log_config.configure_logger(folder_path=folder_path, is_console_logging=True, is_debug_enabled=False)
logger = log_config.singulr_sdk

# read local .env file
from dotenv import load_dotenv, find_dotenv
_ = load_dotenv(find_dotenv())

from typing import (
    Any,
    DefaultDict,
    Dict,
    Optional,
    cast,
)

from singulr_client.common.env_utils import get_runtime_environment


def _serialize_json(obj: Any) -> str:
    """Serialize an object to JSON.

    Parameters
    ----------
    obj : Any
        The object to serialize.

    Returns
    -------
    str
        The serialized JSON string.

    Raises
    ------
    TypeError
        If the object type is not serializable.
    """
    if isinstance(obj, datetime):
        return obj.isoformat()
    else:
        return str(obj)


class LCRunPersister(object):
    """Client for interacting with the Singulr SDK"""

    def create_run(
            self,
            name: str,
            inputs: Dict[str, Any],
            run_type: str,
            *,
            execution_order: Optional[int] = None,
            **kwargs: Any,
    ) -> Optional[Dict[str, str]]:
        """
        Create a new run.

        Parameters
        ----------
        name : str
            The name of the run.
        inputs : Dict[str, Any]
            The inputs for the run.
        run_type : str
            The type of the run.
        execution_order : Optional[int], optional
            The execution order of the run, by default None.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Optional[Dict[str, str]]
            A dictionary with a message indicating the status of the run creation.

        Raises
        ------
        Exception
            If the specified folder location does not exist.
        """
        project_name = kwargs.pop(
            "project_name",
            kwargs.pop(
                "session_name",
                os.environ.get(
                    # TODO: Deprecate LANGCHAIN_SESSION
                    "LANGCHAIN_PROJECT",
                    os.environ.get("LANGCHAIN_SESSION", "default"),
                ),
            ),
        )
        run_create = {
            **kwargs,
            "session_name": project_name,
            "name": name,
            "inputs": inputs,
            "run_type": run_type,
            "execution_order": execution_order if execution_order is not None else 1,
        }
        run_extra = cast(dict, run_create.setdefault("extra", {}))
        runtime = run_extra.setdefault("runtime", {})

        runtime_env = get_runtime_environment()
        run_extra["runtime"] = {**runtime_env, **runtime}
        print("...saving run {}".format(run_create))
        file_name = str(run_create["id"]) + "-" + str(int(time.time()))
        df = pd.DataFrame.from_dict([run_create])
        df = df.astype(str)
        if os.environ.get('LANGCHAIN_RUN_STORAGE_LOCATION'):
            if not os.path.exists(os.environ.get('LANGCHAIN_RUN_STORAGE_LOCATION')):
                logger.info("Folder location does not exist. Environment variable LANGCHAIN_RUN_STORAGE_LOCATION is incorrect.")
                return {"message": "Run created successfully"}
            directory_path = os.environ.get('LANGCHAIN_RUN_STORAGE_LOCATION') + "/create_events"
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)
            df.to_parquet(directory_path + "/{}.parquet".format(file_name), engine='pyarrow')
        return {"message": "Run created successfully"}

    def update_run(
            self,
            run_id: str,
            **kwargs: Any,
    ) -> Optional[Dict[str, str]]:
        """
        Update an existing run.

        Parameters
        ----------
        run_id : str
            The ID of the run to update.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Optional[Dict[str, str]]
            A dictionary with a message indicating the status of the run update.

        Raises
        ------
        Exception
            If the specified folder location does not exist.
        """
        print("...updating run {}".format(run_id))
        file_name = str(kwargs["id"]) + "-" + str(int(time.time()))
        df = pd.DataFrame.from_dict([kwargs])
        df = df.astype(str)
        if os.environ.get('LANGCHAIN_RUN_STORAGE_LOCATION'):
            if not os.path.exists(os.environ.get('LANGCHAIN_RUN_STORAGE_LOCATION')):
                logger.info("Folder location does not exist. Environment variable LANGCHAIN_RUN_STORAGE_LOCATION is incorrect.")
                return {"message": f"Run {run_id} updated successfully"}
            directory_path = os.environ.get('LANGCHAIN_RUN_STORAGE_LOCATION') + "/update_events"
            if not os.path.exists(directory_path):
                os.makedirs(directory_path)
            df.to_parquet(directory_path + "/{}.parquet".format(file_name), engine='pyarrow')
        return {"message": f"Run {run_id} updated successfully"}



# Created by msinghal at 12/09/23
from typing import Optional
from singulr_client.data_contracts.span.primitive_attribute import BaseAttribute, AttributeType


class LLMInvocationParams(BaseAttribute):
    """
    Represents parameters for LLM (Language Model) invocation.

    Attributes:
        INVOCATION_PARAMETERS (str): The key for invocation parameters.
    """

    INVOCATION_PARAMETERS = "invocation_parameters"

    def __init__(self):
        self.subType: str = ""
        self.model: str = ""
        self.modelName: str = ""
        self.requestTimeOutMillis: int = 0
        self.maxRetries: int = 0
        self.maxTokens: Optional[int] = None
        self.stream: bool = False
        self.temperature: float = 0.0

    def get_key(self) -> str:
        """
        Get the key associated with the invocation parameters.

        Returns:
            str: The key for invocation parameters.
        """
        return self.INVOCATION_PARAMETERS

    def get_type(self):
        return AttributeType.LLM_INVOCATION

    def set_key(self, key: str):
        pass

    def set_type(self, type_):
        pass

    def get_sub_type(self) -> str:
        return self.subType

    def set_sub_type(self, subType: str):
        self.subType = subType

    def get_model(self) -> str:
        return self.model

    def set_model(self, model: str):
        self.model = model

    def get_model_name(self) -> str:
        return self.modelName

    def set_model_name(self, modelName: str):
        self.modelName = modelName

    def get_request_time_out_millis(self) -> int:
        return self.requestTimeOutMillis

    def set_request_time_out_millis(self, requestTimeOutMillis: int):
        self.requestTimeOutMillis = requestTimeOutMillis

    def get_max_retries(self) -> int:
        return self.maxRetries

    def set_max_retries(self, maxRetries: int):
        self.maxRetries = maxRetries

    def get_max_tokens(self) -> Optional[int]:
        return self.maxTokens

    def set_max_tokens(self, maxTokens: Optional[int]):
        self.maxTokens = maxTokens

    def is_stream(self) -> bool:
        return self.stream

    def set_stream(self, stream: bool):
        self.stream = stream

    def get_temperature(self) -> float:
        return self.temperature

    def set_temperature(self, temperature: float):
        self.temperature = temperature



import requests
import os
from typing import Optional, Dict, Union

from dotenv import load_dotenv, find_dotenv
_ = load_dotenv(find_dotenv())  # read local .env file


class SingulrClient(object):
    """
    A Python client for sending data to the Singulr API.
    """

    def __init__(self, base_url: Optional[str] = None, api_key: Optional[str] = None):
        """
        Initialize the SingulrClient.

        :param base_url: The base URL of the Singulr API endpoint (default: "http://localhost:8080/api/ingestion/ingest").
        :param api_key: The API key for authentication (optional).
        """
        self.singulr_endpoint = base_url if base_url else os.environ["SINGULR_ENDPOINT"]
        self.api_key = api_key if api_key else None

    def _prepare_headers(self) -> Dict[str, str]:
        """
        Prepare HTTP headers for requests.

        :return: A dictionary of HTTP headers.
        """
        headers = {
            'Content-Type': 'application/json',
        }
        if self.api_key:
            headers['Authorization'] = f'Bearer {self.api_key}'
        return headers

    def ingest_payload(self, ingestion_payload: Dict[str, Union[str, int, float, bool]]) -> Union[requests.Response, None]:
        """
        Send a payload to the Singulr ingestion endpoint.

        :param ingestion_payload: The payload to be ingested as a dictionary.
        :return: The HTTP response object if successful, otherwise None.
        """
        try:
            # Make an HTTP POST request to the Java REST endpoint with the data
            response = requests.post(self.singulr_endpoint, json=ingestion_payload)

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Return the response
                return response
            else:
                raise ValueError(f"HTTP POST request failed with status code: {response.status_code}")

        except requests.exceptions.RequestException as e:
            print(f"An error occurred: {e}")
            return None


if __name__ == '__main__':
    ic = SingulrClient()
    import json

    # Example usage with JSON data
    with open("b.json", "r") as file:
        data = json.load(file)
    ic.ingest_payload(data)



# Created by msinghal at 12/09/23
import time
from typing import Any, Dict


class PayloadMessage(object):
    """
    Represents a payload message.

    Args:
        id (str): The unique identifier for the message.
        payload (Any): The payload data of the message.
    """

    def __init__(self, id: str, payload: Any):
        self.id = id
        self.source = "source-1"
        self.source_type = "LANG_CHAIN"
        self.timestamp = int(time.time() * 1000)
        self.type = "TRACE"
        self.serialization_type = "JSON"
        self.payload = payload

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize the PayloadMessage object to a dictionary.

        Returns:
            Dict[str, Any]: A dictionary representation of the object.
        """
        return {
            "id": self.id,
            "source": self.source,
            "source_type": self.source_type,
            "timestamp": self.timestamp,
            "type": self.type,
            "serialization_type": self.serialization_type,
            "payload": self.payload
        }



# Created by msinghal at 12/09/23
from singulr_client.data_contracts.span.content import KeyedContent
from typing import List, Dict, Any


class Result:
    """
    Represents the result of an operation.

    Attributes:
        inputs (List[KeyedContent]): List of input items.
        outputs (List[KeyedContent]): List of output items.
    """

    def __init__(self):
        self.inputs: List[KeyedContent] = []
        self.outputs: List[KeyedContent] = []

    def get_inputs(self) -> List[KeyedContent]:
        """
        Get the list of input items.

        Returns:
            List[KeyedContent]: List of input items.
        """
        return self.inputs

    def set_inputs(self, inputs: List[KeyedContent]) -> None:
        """
        Set the list of input items.

        Args:
            inputs (List[KeyedContent]): List of input items.
        """
        self.inputs = inputs

    def get_outputs(self) -> List[KeyedContent]:
        """
        Get the list of output items.

        Returns:
            List[KeyedContent]: List of output items.
        """
        return self.outputs

    def set_outputs(self, outputs: List[KeyedContent]) -> None:
        """
        Set the list of output items.

        Args:
            outputs (List[KeyedContent]): List of output items.
        """
        self.outputs = outputs

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the result to a dictionary.

        Returns:
            Dict[str, Any]: The result represented as a dictionary.
        """
        result_dict = {
            "inputs": [input_item.to_dict() for input_item in self.inputs],
            "outputs": [output_item.to_dict() for output_item in self.outputs]
        }
        return result_dict



# Importing necessary modules
# Created by msinghal at 12/09/23
import platform  # Module for platform information
import os
import socket
import psutil  # Modules for OS, networking, and process information
from functools import lru_cache  # Importing LRU cache decorator

def get_memory() -> int:
    """
    Get the total available memory in bytes.

    Returns:
        int: Total available memory in bytes.
    """
    try:
        memory_info = psutil.virtual_memory()
        total_memory = memory_info.total
    except ImportError:
        # If psutil is not available, attempt to calculate total memory using os.sysconf
        total_memory = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
    return total_memory

def get_process() -> dict:
    """
    Get information about the current process.

    Returns:
        dict: A dictionary containing information about the current process.
            - 'process_id': Process ID (PID) as an integer.
            - 'process_name': Name of the process as a string.
            - 'process_status': Status of the process as a string.
            - 'process_start_time': Process start time in seconds since the epoch as a float.
    """
    current_pid = os.getpid()
    current_process = psutil.Process(current_pid)
    return {
        "process_id": current_pid,
        "process_name": str(current_process.cmdline()[-1]),
        "process_status": current_process.status(),
        "process_start_time": current_process.create_time()
    }

@lru_cache(maxsize=1)  # Using LRU cache with a maximum size of 1
def get_runtime_environment() -> dict:
    """
    Get information about the runtime environment.

    Returns:
        dict: A dictionary containing information about the runtime environment.
            - 'library_version': Version of the 'langchain' library as a string.
            - 'langchain_version': Version of 'langchain' as a string (assuming it's imported from the module).
            - 'library': Name of the library as a string.
            - 'platform': Platform information as a string.
            - 'runtime': Name of the runtime as a string ('python').
            - 'runtime_version': Version of Python as a string.
            - 'platform_architecture': Platform architecture as a string.
            - 'platform_node': Node name as a string.
            - 'platform_version': Platform version as a string.
            - 'platform_cpu': Number of CPUs as an integer.
            - 'platform_memory': Total available memory in bytes as an integer.
            - 'node_name': Hostname as a string.
            - 'node_ip': IP address of the hostname as a string.
            - 'node_fqdn': Fully qualified domain name as a string.
            - 'process_name': Name of the current process as a string.
            - 'process_id': Process ID (PID) of the current process as an integer.
            - 'process_status': Status of the current process as a string.
            - 'process_start_time': Start time of the current process in seconds since the epoch as a float.
    """
    # Lazy import to avoid circular imports
    from langchain import __version__  # Assuming the version information is imported from langchain module
    current_process_info = get_process()
    return {
        "library_version": __version__,
        "langchain_version": __version__,
        "library": "langchain",
        "platform": platform.platform(),
        "runtime": "python",
        "runtime_version": platform.python_version(),
        "platform_architecture": ''.join(platform.architecture()),
        "platform_node": platform.node(),
        "platform_version": platform.version(),
        "platform_cpu": os.cpu_count(),
        "platform_memory": get_memory(),
        "node_name": socket.gethostname(),
        "node_ip": "127.0.0.1",
        "node_fqdn": socket.getfqdn(),
        "process_name": current_process_info["process_name"],
        "process_id": current_process_info["process_id"],
        "process_status": current_process_info["process_status"],
        "process_start_time": current_process_info["process_start_time"],
    }

if __name__ == '__main__':
    get_runtime_environment()



# Created by msinghal at 04/10/23
from setuptools import setup, find_packages

setup(
    name='singulr',
    version='0.0.9',
    description='Package to trace generative apps and guard them',
    url='',
    author='Abhijit Sharma',
    author_email='abhijit@singulr.com',
    license='MIT',  # Choose an appropriate license

    # Specify the packages to include
    packages=find_packages(),

    # Define package dependencies
    install_requires=[
        'langchain'
        # List your dependencies here
    ],
)


# Created by msinghal at 12/09/23
from enum import Enum
from typing import List, Optional, Dict, Any
from singulr_client.data_contracts.span.primitive_attribute import PrimitiveAttribute
from singulr_client.data_contracts.span.result import Result


class SpanType(Enum):
    """
    Enumeration representing different types of spans.
    """
    UNKNOWN = "UNKNOWN"
    LLM = "LLM"
    CHAIN = "CHAIN"
    TOOL = "TOOL"
    AGENT = "AGENT"
    RETRIEVER = "RETRIEVER"

    def __str__(self) -> str:
        return str(self.value)


class StatusCode(str, Enum):
    """
    Enumeration representing status codes.
    """
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)


class Span(object):
    """
    Represents a span.

    Args:
        span_id (Optional[str]): The unique identifier of the span.
        name (Optional[str]): The name of the span.
        parent_span_id (Optional[str]): The ID of the parent span.
        trace_id (Optional[str]): The ID of the trace.
        start_time_millis (Optional[int]): The start time of the span in milliseconds.
        end_time_millis (Optional[int]): The end time of the span in milliseconds.
        status_code (Optional[StatusCode]): The status code of the span.
        status_message (Optional[str]): The status message of the span.
        type (Optional[SpanType]): The type of the span.
        sub_type (Optional[str]): The sub-type of the span.
        attributes (Optional[List[PrimitiveAttribute]]): List of attributes associated with the span.
        results (Optional[List[Result]]): List of results associated with the span.
        child_spans (Optional[List["Span"]]): List of child spans.

    Attributes:
        span_id (Optional[str]): The unique identifier of the span.
        name (Optional[str]): The name of the span.
        parent_span_id (Optional[str]): The ID of the parent span.
        trace_id (Optional[str]): The ID of the trace.
        start_time_millis (Optional[int]): The start time of the span in milliseconds.
        end_time_millis (Optional[int]): The end time of the span in milliseconds.
        status_code (Optional[StatusCode]): The status code of the span.
        status_message (Optional[str]): The status message of the span.
        type (Optional[SpanType]): The type of the span.
        sub_type (Optional[str]): The sub-type of the span.
        attributes (List[PrimitiveAttribute]): List of attributes associated with the span.
        results (List[Result]): List of results associated with the span.
        child_spans (List["Span"]): List of child spans.
    """

    def __init__(
            self,
            span_id: Optional[str] = None,
            name: Optional[str] = None,
            parent_span_id: Optional[str] = None,
            trace_id: Optional[str] = None,
            start_time_millis: Optional[int] = None,
            end_time_millis: Optional[int] = None,
            status_code: Optional[StatusCode] = None,
            status_message: Optional[str] = None,
            type: Optional[SpanType] = None,
            sub_type: Optional[str] = None,
            attributes: Optional[List[PrimitiveAttribute]] = None,
            results: Optional[List[Result]] = None,
            child_spans: Optional[List["Span"]] = None,
    ):
        self.span_id = span_id
        self.name = name
        self.parent_span_id = parent_span_id
        self.trace_id = trace_id
        self.start_time_millis = start_time_millis
        self.end_time_millis = end_time_millis
        self.status_code = status_code
        self.status_message = status_message
        self.type = type
        self.sub_type = sub_type
        self.attributes = attributes if attributes is not None else []
        self.results = results if results is not None else []
        self.child_spans = child_spans if child_spans is not None else []

    def add_attribute(self, key: str, value: Any, at_type: str, at_value: str) -> None:
        """
        Add an attribute to the span.

        Args:
            key (str): The key of the attribute.
            value (Any): The value of the attribute.
            at_type (str): The type of the attribute.
            at_value (str): The value of the attribute.
        """
        attr = {
            "@type": at_type,
            "key": key,
            "type": at_value,
            "value": value
        }
        return attr

    def add_named_result(self, inputs: Dict[str, Any], outputs: Dict[str, Any]) -> None:
        """
        Add a named result to the span.

        Args:
            inputs (Dict[str, Any]): The input data.
            outputs (Dict[str, Any]): The output data.
        """
        if self.results is None:
            self.results = []
        self.results.append(Result(inputs, outputs))

    def add_child_span(self, span: "Span") -> None:
        """
        Add a child span to the span.

        Args:
            span (Span): The child span to be added.
        """
        if self.child_spans is None:
            self.child_spans = []
        self.child_spans.append(span)

    def get_span_id(self) -> Optional[str]:
        """
        Get the span ID.

        Returns:
            Optional[str]: The span ID.
        """
        return self.span_id

    def set_span_id(self, span_id: str) -> None:
        """
        Set the span ID.

        Args:
            span_id (str): The span ID.
        """
        self.span_id = span_id

    def get_name(self) -> Optional[str]:
        """
        Get the name of the span.

        Returns:
            Optional[str]: The name of the span.
        """
        return self.name

    def set_name(self, name: str) -> None:
        """
        Set the name of the span.

        Args:
            name (str): The name of the span.
        """
        self.name = name

    # ... (similar methods for other attributes)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the span to a dictionary.

        Returns:
            Dict[str, Any]: The span represented as a dictionary.
        """
        span_dict = {
            "span_id": self.span_id,
            "name": self.name,
            "parent_span_id": self.parent_span_id,
            "trace_id": self.trace_id,
            "start_time_millis": self.start_time_millis,
            "end_time_millis": self.end_time_millis,
            "status_code": self.status_code.value,
            "status_message": self.status_message,
            "type": self.type.value if self.type else SpanType.UNKNOWN.value,
            "sub_type": self.sub_type,
            "attributes": [attr.to_dict() for attr in self.attributes],
            "results": [result.to_dict() for result in self.results],
            "child_spans": [span.to_dict() for span in self.child_spans],
        }
        # Remove fields with None values
        span_dict = {k: v for k, v in span_dict.items() if v is not None}
        return span_dict



# Created by msinghal at 12/09/23
from singulr_client.data_contracts.span.primitive_attribute import BaseAttribute, AttributeType

class TokenUsage:
    """
    Represents token usage statistics.

    Attributes:
        promptTokens (int): The number of prompt tokens used.
        completionTokens (int): The number of completion tokens used.
        totalTokens (int): The total number of tokens used.
    """

    def __init__(self):
        self.promptTokens: int = 0
        self.completionTokens: int = 0
        self.totalTokens: int = 0

    def get_prompt_tokens(self) -> int:
        """
        Get the number of prompt tokens used.

        Returns:
            int: The number of prompt tokens used.
        """
        return self.promptTokens

    def set_prompt_tokens(self, promptTokens: int):
        """
        Set the number of prompt tokens used.

        Args:
            promptTokens (int): The number of prompt tokens used.
        """
        self.promptTokens = promptTokens

    def get_completion_tokens(self) -> int:
        """
        Get the number of completion tokens used.

        Returns:
            int: The number of completion tokens used.
        """
        return self.completionTokens

    def set_completion_tokens(self, completionTokens: int):
        """
        Set the number of completion tokens used.

        Args:
            completionTokens (int): The number of completion tokens used.
        """
        self.completionTokens = completionTokens

    def get_total_tokens(self) -> int:
        """
        Get the total number of tokens used.

        Returns:
            int: The total number of tokens used.
        """
        return self.totalTokens

    def set_total_tokens(self, totalTokens: int):
        """
        Set the total number of tokens used.

        Args:
            totalTokens (int): The total number of tokens used.
        """
        self.totalTokens = totalTokens

class LLMStats(BaseAttribute):
    """
    Represents Language Model (LLM) statistics.

    Attributes:
        LLM_OUTPUT (str): The key for LLM output.
    """

    LLM_OUTPUT = "llm_output"

    def __init__(self):
        self.subType: str = ""
        self.tokenUsage: TokenUsage = TokenUsage()
        self.modelName: str = ""

    def get_key(self) -> str:
        """
        Get the key associated with LLM output.

        Returns:
            str: The key for LLM output.
        """
        return self.LLM_OUTPUT

    def get_type(self):
        return AttributeType.LLM_STATS

    def set_key(self, key: str):
        pass

    def set_type(self, type_):
        pass

    def get_sub_type(self) -> str:
        return self.subType

    def set_sub_type(self, subType: str):
        """
        Set the subtype of LLM statistics.

        Args:
            subType (str): The subtype of LLM statistics.
        """
        self.subType = subType

    def get_token_usage(self) -> TokenUsage:
        """
        Get the token usage statistics.

        Returns:
            TokenUsage: Token usage statistics.
        """
        return self.tokenUsage

    def set_token_usage(self, tokenUsage: TokenUsage):
        """
        Set the token usage statistics.

        Args:
            tokenUsage (TokenUsage): Token usage statistics.
        """
        self.tokenUsage = tokenUsage

    def get_model_name(self) -> str:
        """
        Get the name of the language model.

        Returns:
            str: The name of the language model.
        """
        return self.modelName

    def set_model_name(self, modelName: str):
        """
        Set the name of the language model.

        Args:
            modelName (str): The name of the language model.
        """
        self.modelName = modelName



# # Created by msinghal at 11/09/23
# import typing, dataclasses
# from singulr_client.span.span import Span
# from singulr_client.env.environment import Environment
# from singulr_client.utils import _safe_serialize, _hash_id
# from langchain.callbacks.tracers.schemas import Run
# from singulr_client.trace.trace import TraceTree
# from dataclasses import dataclass, field
#
#
#
# class SingulrTraceTree():
#     """Media object for trace tree data.
#
#     Arguments:
#         root_span (Span): The root span of the trace tree.
#         model_dict (dict, optional): A dictionary containing the model dump.
#             NOTE: model_dict is a completely-user-defined dict. The UI will render
#             a JSON viewer for this dict, giving special treatment to dictionaries
#             with a `_kind` key. This is because model vendors have such different
#             serialization formats that we need to be flexible here.
#     """
#
#     def __init__(
#             self,
#             root_span: Span,
#             model_dict: typing.Optional[dict] = None,
#             environment_info: Environment = None,
#             trace_id:str=None
#     ):
#         super().__init__()
#         self._root_span = root_span
#         self._model_dict = model_dict
#         self._environment_info = environment_info
#         self._trace_id = trace_id
#
#     def to_json(self) -> dict:
#
#         span_dict = self._root_span.to_dict()
#         env_dict = self._environment_info.to_dict()
#         environment = _safe_serialize(dataclasses.asdict(self._environment_info))
#         trace_tree = {"trace_id":  self._trace_id, "root_span": res["root_span_dumps"], "environment": environment}
#         print("trace_tree: \n {}".format(trace_tree))
#         return trace_tree
#
#     def is_bound(self) -> bool:
#         return True
#
#
